import io
import logging
from datetime import timedelta

from astropy.io import fits
from dateutil.parser import parse

from lco_ingester.exceptions import DoNotRetryError
from lco_ingester.utils.fits import reduction_level, related_for_catalog
from lco_ingester.utils.fits import File
from lco_ingester.settings import settings

logger = logging.getLogger('lco_ingester')


class FitsDict(object):
    INTEGER_TYPES = ['BLKUID', 'REQNUM']

    def __init__(self, file, required_headers, blacklist_headers):
        self.file = file
        self.required_headers = required_headers
        self.blacklist_headers = list(blacklist_headers)
        # Make sure to blacklist empty headers, since we never want them and they cause problems.
        if '' not in self.blacklist_headers:
            self.blacklist_headers.append('')

    def get_hdu_with_required_headers(self):
        with self.file.get_fits() as fits_file:
            with fits.open(io.BytesIO(fits_file.read()), mode='readonly') as hdulist:
                for hdu in hdulist:
                    fits_dict = dict(hdu.header)
                    if any([k for k in self.required_headers if k not in fits_dict]):
                        pass
                    else:
                        self.fits_dict = fits_dict
                        logger.info('Ingester extracted fits headers', extra={
                            'tags': {
                                'request_num': fits_dict.get('REQNUM'),
                                'PROPID': fits_dict.get('PROPID'),
                                'filename': '{}{}'.format(self.file.basename, self.file.extension)
                            }
                        })
                        return
                raise DoNotRetryError('Could not find required headers!')  # No headers met requirement

    def remove_blacklist_headers(self):
        for header in self.blacklist_headers:
            try:
                del(self.fits_dict[header])
            except KeyError:
                pass

    def normalize_null_values(self):
        #  Sometimes keywords use N/A to mean null
        for k, v in self.fits_dict.items():
            if v == 'N/A' or v == 'UNKNOWN' or v == 'UNSPECIFIED':
                if k in self.INTEGER_TYPES:
                    self.fits_dict[k] = None
                else:
                    self.fits_dict[k] = ''

    def check_rlevel(self):
        if not self.fits_dict.get('RLEVEL'):
            # Check if the frame contains its reduction level, if not deduce it
            rlevel = reduction_level(self.file.basename, self.file.extension)
            self.fits_dict['RLEVEL'] = rlevel

    def check_catalog(self):
        if '_cat' in self.file.basename:
            if not self.fits_dict.get('L1IDCAT'):
                # Check if the catalog file contains it's target frame, if not deduce it
                l1idcat = related_for_catalog(self.file.basename)
                self.fits_dict['L1IDCAT'] = l1idcat
            # set obstype to CATALOG even though it's set to EXPOSE by the pipeline
            self.fits_dict['OBSTYPE'] = 'CATALOG'

    def set_public_date(self):
        if not self.fits_dict.get('L1PUBDAT'):
            # Check if the frame doesnt specify a public date.
            if (self.fits_dict['OBSTYPE'] in settings.CALIBRATION_TYPES or
                    any([prop in self.fits_dict['PROPID'] for prop in settings.PUBLIC_PROPOSALS])):
                self.fits_dict['L1PUBDAT'] = self.fits_dict['DATE-OBS']
            else:
                # This should be proprietary, set it to a year from DATE-OBS
                self.fits_dict['L1PUBDAT'] = (
                    parse(self.fits_dict['DATE-OBS']) + timedelta(days=365)
                ).isoformat()

    def normalize_related(self):
        """
        Fits files contain several keys whose values are the filenames
        of frames that they are related to.
        Sometimes the keys are non-existant, sometimes they contain
        'N/A' to represent null, sometimes they contain filenames with
        the file extension appended, sometimes without an extension.
        This function attempts to normalize these values to not exist
        if the value should be null, or if they do, be the filename
        without extension.
        """
        related_frame_keys = [
            'L1IDBIAS', 'L1IDDARK', 'L1IDFLAT', 'L1IDSHUT',
            'L1IDMASK', 'L1IDFRNG', 'L1IDCAT', 'TARFILE',
            'ORIGNAME', 'ARCFILE', 'FLATFILE', 'GUIDETAR'
        ]
        for key in related_frame_keys:
            filename = self.fits_dict.get(key)
            if filename and filename != 'N/A':
                # The key has a value that isn't NA, we have a related frame
                basename, extension = File.get_basename_and_extension(filename)
                if extension:
                    # This value has an extention, so remove it.
                    self.fits_dict[key] = basename
                else:
                    pass
            else:
                # If the value is NA or the key doesn't exit, make sure it
                # is not the in dictionary we return
                try:
                    del self.fits_dict[key]
                except KeyError:
                    pass

    def as_dict(self):
        self.get_hdu_with_required_headers()
        self.remove_blacklist_headers()
        self.normalize_null_values()
        self.check_rlevel()
        self.check_catalog()
        self.set_public_date()
        self.normalize_related()
        return self.fits_dict



from kombu.connection import Connection
from kombu import Exchange, Queue

from lco_ingester.settings import settings

processed_exchange = Exchange(settings.PROCESSED_EXCHANGE_NAME, type='fanout')
producer_queue = Queue('', processed_exchange, exclusive=True)


class PostProcService(object):
    def __init__(self, broker_url):
        self.broker_url = broker_url

    def post_to_archived_queue(self, fits_dict):
        with Connection(self.broker_url) as conn:
            queue = conn.SimpleQueue(producer_queue)
            queue.put(fits_dict)
            queue.close()



import logging
from datetime import datetime

import requests
from opentsdb_python_metrics.metric_wrappers import metric_timer, SendMetricMixin

from lco_ingester.postproc import PostProcService
from lco_ingester.utils.fits import obs_end_time_from_dict
from lco_ingester.exceptions import BackoffRetryError, RetryError

logger = logging.getLogger('lco_ingester')


class ArchiveService(SendMetricMixin):
    def __init__(self, api_root, auth_token, broker_url):
        self.api_root = api_root
        self.headers = {'Authorization': 'Token {}'.format(auth_token)}
        self.broker_url = broker_url

    def handle_response(self, response):
        try:
            response.raise_for_status()
        except requests.exceptions.ConnectionError as exc:
            raise BackoffRetryError(exc)
        except requests.exceptions.HTTPError as exc:
            raise RetryError(exc)
        return response.json()

    def version_exists(self, md5):
        response = requests.get(
            '{0}versions/?md5={1}'.format(self.api_root, md5), headers=self.headers
        )
        result = self.handle_response(response)
        try:
            return result['count'] > 0
        except KeyError as e:
            raise BackoffRetryError(e)

    @metric_timer('ingester.post_frame')
    def post_frame(self, fits_dict):
        response = requests.post(
            '{0}frames/'.format(self.api_root), json=fits_dict, headers=self.headers
        )
        result = self.handle_response(response)
        logger.info('Ingester posted frame to archive', extra={
            'tags': {
                'filename': result.get('filename'),
                'request_num': fits_dict.get('REQNUM'),
                'PROPID': result.get('PROPID'),
                'id': result.get('id')
            }
        })
        # Add some useful information from the result
        fits_dict['frameid'] = result.get('id')
        fits_dict['filename'] = result.get('filename')
        fits_dict['url'] = result.get('url')
        # Send frame that was posted to the fits exchange
        post_proc = PostProcService(self.broker_url)
        post_proc.post_to_archived_queue(fits_dict)
        # Record metric for the ingest lag (time between date of image vs date ingested)
        ingest_lag = datetime.utcnow() - obs_end_time_from_dict(fits_dict)
        self.send_metric('ingester.ingest_lag', ingest_lag.total_seconds())
        return fits_dict



import os
from datetime import timedelta

from lco_ingester.settings import settings


# Celery Broker
broker_url = os.getenv('BROKER_URL', 'memory://localhost')

# Celery Settings
celery_config = {
    'broker_url': broker_url,
    'task_soft_time_limit': 1200,
    'task_time_limit': 3600,
    'worker_prefetch_multiplier': 1,
    'worker_max_tasks_per_child': 10,
    'beat_schedule': {
        'queue-length-every-minute': {
            'task': 'tasks.collect_queue_length_metric',
            'schedule': timedelta(minutes=1),
            'args': ('http://ingesterrabbitmq:15672/',),
            'options': {'queue': 'periodic'}
        },
        'total-holdings-every-5-minutes': {
            'task': 'tasks.total_holdings',
            'schedule': timedelta(minutes=5),
            'args': (settings.API_ROOT, settings.AUTH_TOKEN),
            'options': {'queue': 'periodic'}
        }
    }
}



import hashlib
import logging
from io import BytesIO
from datetime import datetime

from lco_ingester.utils.fits import get_storage_class
from opentsdb_python_metrics.metric_wrappers import metric_timer, SendMetricMixin
from botocore.exceptions import EndpointConnectionError, ConnectionClosedError

import requests
import boto3

from lco_ingester.exceptions import BackoffRetryError

logger = logging.getLogger('lco_ingester')


class S3Service(SendMetricMixin):
    def __init__(self, bucket):
        self.bucket = bucket

    def file_to_s3_key(self, file, fits_dict):
        site = fits_dict.get('SITEID')
        instrument = fits_dict.get('INSTRUME')
        date_obs = fits_dict.get('DATE-OBS')
        if site.lower() == 'sor':
            # SOR files don't have the day_obs in their filename, so use the DATE_OBS field:
            day_obs = date_obs.split('T')[0].replace('-', '')
        else:
            day_obs = file.basename.split('-')[2]
        return '/'.join((site, instrument, day_obs, file.basename)) + file.extension

    def extension_to_content_type(self, extension):
        content_types = {
            '.fits': 'image/fits',
            '.tar.gz': 'application/x-tar',
        }
        return content_types.get(extension, '')

    def strip_quotes_from_etag(self, etag):
        """
        Amazon returns the md5 sum of the uploaded
        file in the 'ETag' header wrapped in quotes
        """
        if etag.startswith('"') and etag.endswith('"'):
            return etag[1:-1]

    @metric_timer('ingester.upload_file')
    def upload_file(self, file, fits_dict):
        storage_class = get_storage_class(fits_dict)
        start_time = datetime.utcnow()
        s3 = boto3.resource('s3')
        key = self.file_to_s3_key(file, fits_dict)
        content_disposition = 'attachment; filename={0}{1}'.format(file.basename, file.extension)
        content_type = self.extension_to_content_type(file.extension)
        try:
            response = s3.Object(self.bucket, key).put(
                Body=file.get_from_start(),
                ContentDisposition=content_disposition,
                ContentType=content_type,
                StorageClass=storage_class,
            )
        except (requests.exceptions.ConnectionError,
                EndpointConnectionError, ConnectionClosedError) as exc:
            raise BackoffRetryError(exc)
        s3_md5 = self.strip_quotes_from_etag(response['ETag'])
        key = response['VersionId']
        logger.info('Ingester uploaded file to s3', extra={
            'tags': {
                'filename': '{}{}'.format(file.basename, file.extension),
                'key': key,
                'storage_class': storage_class,
            }
        })
        # Record metric for the bytes transferred / time to upload
        upload_time = datetime.utcnow() - start_time
        bytes_per_second = len(file) / upload_time.total_seconds()
        self.send_metric('ingester.s3_upload_bytes_per_second', bytes_per_second)
        # TODO: Remove 'migrated': True from the return dict when the s3 migration is complete
        return {'key': key, 'md5': s3_md5, 'extension': file.extension, 'migrated': True}

    @staticmethod
    def get_file(path):
        """
        Get a file from s3.

        Path is of the pattern `s3://{bucket}/{s3_key}`

        :param path: Path to file in s3
        :return: File-like object
        """
        s3 = boto3.resource('s3')
        protocol_preface = 's3://'
        plist = path[len(protocol_preface):].split('/')
        bucket = plist[0]
        key = '/'.join(plist[1:])
        o = s3.Object(key=key, bucket_name=bucket).get()
        filename = o['ContentDisposition'].split('filename=')[1]
        fileobj = BytesIO(o['Body'].read())
        fileobj.name = filename
        return fileobj



from lco_ingester.fits import FitsDict
from lco_ingester.exceptions import BackoffRetryError, NonFatalDoNotRetryError
from lco_ingester.utils.fits import wcs_corners_from_dict
from lco_ingester.utils.fits import File
from lco_ingester.archive import ArchiveService
from lco_ingester.s3 import S3Service
from lco_ingester.settings import settings


def frame_exists(fileobj, api_root=settings.API_ROOT, auth_token=settings.AUTH_TOKEN, broker_url=settings.FITS_BROKER):
    """
    Checks if the frame exists in the archive.

    :param fileobj: File-like object
    :param api_root: Archive API root url
    :param auth_token: Archive API authentication token
    :return: Boolean indicating whether the frame exists
    """
    archive = ArchiveService(api_root=api_root, auth_token=auth_token, broker_url=broker_url)
    md5 = File(fileobj, run_validate=False).get_md5()
    return archive.version_exists(md5)


def validate_fits_and_create_archive_record(fileobj, path=None, required_headers=settings.REQUIRED_HEADERS,
                                            blacklist_headers=settings.HEADER_BLACKLIST):
    """
    Validate the fits file and also create an archive record from it.

    :param fileobj: File-like object
    :param path: File path/name for this object
    :param required_headers: FITS headers that must be present
    :param blacklist_headers: FITS headers that should not be ingested
    :return: Constructed archive record
    """
    file = File(fileobj, path)
    json_record = FitsDict(file, required_headers, blacklist_headers).as_dict()
    json_record['area'] = wcs_corners_from_dict(json_record)
    json_record['basename'] = file.basename
    return json_record


def upload_file_to_s3(fileobj, path=None, bucket=settings.BUCKET):
    """
    Uploads a file to s3.

    :param fileobj: File-like object
    :param path: File path/name for this object
    :param bucket: S3 bucket name
    :return: Version information for the file that was uploaded
    """
    file = File(fileobj, path)
    s3 = S3Service(bucket)

    # Transform this fits file into a cleaned dictionary
    fits_dict = FitsDict(file, settings.REQUIRED_HEADERS, settings.HEADER_BLACKLIST).as_dict()

    # Returns the version, which holds in it the md5 that was uploaded
    return s3.upload_file(file, fits_dict)


def ingest_archive_record(version, record, api_root=settings.API_ROOT, auth_token=settings.AUTH_TOKEN,
                          broker_url=settings.FITS_BROKER):
    """
    Ingest an archive record.

    :param version: Result of the upload to s3
    :param record: Archive record to ingest
    :param api_root: Archive API root url
    :param auth_token: Archive API authentication token
    :param broker_url: FITS exchange broker
    :return: The archive record that was ingested
    """
    archive = ArchiveService(api_root=api_root, auth_token=auth_token, broker_url=broker_url)
    # Construct final archive payload and post to archive
    record['version_set'] = [version]
    return archive.post_frame(record)


def upload_file_and_ingest_to_archive(fileobj, path=None, required_headers=settings.REQUIRED_HEADERS,
                                      blacklist_headers=settings.HEADER_BLACKLIST,
                                      api_root=settings.API_ROOT, auth_token=settings.AUTH_TOKEN,
                                      bucket=settings.BUCKET, broker_url=settings.FITS_BROKER):
    """
    Ingest and upload a file.

    :param fileobj: File-like object
    :param path: File path/name for this object
    :param api_root: Archive API root url
    :param auth_token: Archive API authentication token
    :param bucket: S3 bucket name
    :param required_headers: FITS headers that must be present
    :param blacklist_headers: FITS headers that should not be ingested
    :param broker_url: FITS exchange broker
    :return: Information about the uploaded file and record
    """
    file = File(fileobj, path)
    archive = ArchiveService(api_root=api_root, auth_token=auth_token, broker_url=broker_url)
    s3 = S3Service(bucket)
    ingester = Ingester(file, s3, archive, required_headers, blacklist_headers)
    return ingester.ingest()


class Ingester(object):
    """
    Ingest a single file into the archive.

    A single instance of this class is responsible for parsing a fits file,
    uploading the data to s3, and making a call to the archive api.
    """
    def __init__(self, file, s3, archive, required_headers=None, blacklist_headers=None):
        self.file = file
        self.s3 = s3
        self.archive = archive
        self.required_headers = required_headers if required_headers else []
        self.blacklist_headers = blacklist_headers if blacklist_headers else []

    def ingest(self):
        # Get the Md5 checksum of this file and check if it already exists in the archive
        md5 = self.file.get_md5()
        if self.archive.version_exists(md5):
            raise NonFatalDoNotRetryError('Version with this md5 already exists')

        # Transform this fits file into a cleaned dictionary
        fits_dict = FitsDict(self.file, self.required_headers, self.blacklist_headers).as_dict()

        # Upload the file to s3 and get version information back
        version = self.s3.upload_file(self.file, fits_dict)

        # Make sure our md5 matches amazons
        if version['md5'] != md5:
            raise BackoffRetryError('S3 md5 did not match ours')

        # Construct final archive payload and post to archive
        fits_dict['area'] = wcs_corners_from_dict(fits_dict)
        fits_dict['version_set'] = [version]
        fits_dict['basename'] = self.file.basename
        return self.archive.post_frame(fits_dict)
